// H.A.L.O. Project
// Haptic Assisted Locating of Objects
// steve@polymythic.com
// www.polymythic.com
// Steve Struebing
// 12_2_2010 - Changing the motor scheme
// 12_10_2010 - Changed pulsein in ping library 
// 12_10_2010_v6- using analogWrite() for intensity rather than digital out
// 12_10_2010_v7- modifying range algorithm

// ****************************************** 
// INCLUDES 
//http://www.parallax.com/dl/docs/prod/acc/28015-PING-v1.3.pdf
//
#include <Ping2.h>  // Modified Parallax Ping Library

// ****************************************** 
// CALIBRATION VARIABLES
// INTERPULSE_LANTENCY (milliseconds) - Defines the period of time between when one ping
// fires and the next fires.  This should allow for the sound to travel to
// its furthest detectable echo distance so that the next sensors RX will not
// pick up splash from the previous ping's TX.  This may/will result in 
// a lower than real distance result from the next sensor.
#define INTERPULSE_LATENCY_DURATION 25

// MOTOR_PULSE_DURATION (milliseconds) - Defines the duration that the motor will pulse
// This is a constant at the moment, and the MOTOR_VIBRATION_FACTOR will
// shorte between the pulses of the motor
#define MOTOR_PULSE_DURATION 250

// RESCHEDULE_THRESHOLD_CONSTANT - Note: This may be replaced by "debouncing?" the motor
// This is the threshold over which the distance is large enough to trigger a 
// reschedule event.  If too small, there may be a lot of rescheduling, and if
// it is too large, then there may not be enough granularity.
//#define RESCHEDULE_THRESHOLD_CONSTANT 20.83

// MAXIMUM_DISTANCE_CONSTANT (inches) - This is is the range over which we will
// not schedule any events.  If the range is greater than this value, we will not even
// bother with it.
#define MAXIMUM_DISTANCE_CONSTANT 48

//#define INTENSITY_STEP 63

// MOTOR_DISTANCE_GAP_FACTOR (millis per inch) - 1000 milis / MAXIMUM_DISTANCE_CONSTANT = milis per inch
// NOTE: This is an important constant.  If the value is too low, then 
// the motors will constantly be firing and rescheduling the next start
// which may elapse before any motor stop happens.  This will result
// in always on motors.
// This was calculated at distance of 48 inches (1000 / 48)
// @todo- this should be derived from the MAXIMUM_DISTANCE_CONSTANT
//#define MOTOR_DISTANCE_GAP_FACTOR 20.83
#define MOTOR_DISTANCE_GAP_FACTOR 250

#define MAX_GAP ((MAXIMUM_DISTANCE_CONSTANT/12) * MOTOR_DISTANCE_GAP_FACTOR)

// PING_MS_DELAY - THIS HAS BEEN MOVED INTO THE PING2 LIBRARY
// Note: This caused some debugging because the pulseIn() command used in the normal
// library defaults to 1 second.  This is way outside what we can handle.  According
// to the PING)) datasheet, the maximum time to wait (tIN_MAX) == 115us.  We will use
// 150 just for some buffer.
//#define PING_MS_DELAY 150

// ****************************************** 
// INPUT PINS 
// - PING Input Sensors 
#define INPUT_LEFT_SENSOR_PIN         3
#define INPUT_LEFT_CENTER_SENSOR_PIN  4
#define INPUT_CENTER_SENSOR_PIN       5 
#define INPUT_RIGHT_CENTER_SENSOR_PIN 6
#define INPUT_RIGHT_SENSOR_PIN        7

// Enumeration for sensor input array
enum sensor_input {
  left_sensor,
  left_center_sensor,
  center_sensor,
  right_center_sensor,
  right_sensor,
  max_sensor_index
};

// ****************************************** 
// OUTPUT PINS
// - MOTOR Output Pins (PWM)
#define OUTPUT_LEFT_MOTOR_PIN         12
#define OUTPUT_LEFT_CENTER_MOTOR_PIN  11         
#define OUTPUT_CENTER_MOTOR_PIN       10
#define OUTPUT_RIGHT_CENTER_MOTOR_PIN 9
#define OUTPUT_RIGHT_MOTOR_PIN        8
#define OUTPUT_DEBUG_PIN              13
/*
// - DEBUG Output Pin (LED)
#define OUTPUT_CENTER_MOTOR_PIN       13
*/
// Enumeration for motor output array
enum motor_output {
  left_motor,
  left_center_motor,
  center_motor,
  right_center_motor,
  right_motor,
  max_motor_index
};

// Create the array of input PING))) sensors
Ping2 inputSensorArray[max_sensor_index] = {INPUT_LEFT_SENSOR_PIN, 
                                            INPUT_LEFT_CENTER_SENSOR_PIN,
                                            INPUT_CENTER_SENSOR_PIN,
                                            INPUT_RIGHT_CENTER_SENSOR_PIN,
                                            INPUT_RIGHT_SENSOR_PIN};                                        

// Create the array of output vibration motors based on pin constants selected
int outputMotorArrayPins[max_motor_index] = {OUTPUT_LEFT_MOTOR_PIN, 
                                           OUTPUT_LEFT_CENTER_MOTOR_PIN,
                                           OUTPUT_CENTER_MOTOR_PIN, 
                                           OUTPUT_RIGHT_CENTER_MOTOR_PIN,
                                           OUTPUT_RIGHT_MOTOR_PIN};                                         

// Create the array of input sensors based on pin constants selected
int inputSensorArrayPins[max_sensor_index] = {INPUT_LEFT_SENSOR_PIN, 
                                           INPUT_LEFT_CENTER_SENSOR_PIN,
                                           INPUT_CENTER_SENSOR_PIN, 
                                           INPUT_RIGHT_CENTER_SENSOR_PIN,
                                           INPUT_RIGHT_SENSOR_PIN};

// Create the schedule data members
unsigned long motorStartTimeSchedule[max_motor_index];  // this is the clock time (determined by millis() ) when we should start a motor
unsigned long motorStartNextTimeSchedule[max_motor_index];  // this is the clock time (determined by millis() ) when we should start a motor
unsigned long motorStopTimeSchedule[max_motor_index];   // this is the clock time (determines by millis() ) when we should stop a motor 
unsigned long motorStopNextTimeSchedule[max_motor_index];  // this is the clock time (determined by millis() ) when we should start a motor
unsigned int  motorStopTimeGap[max_motor_index];        // this is the gap of time (in milliseconds) between a motor stop and start
unsigned int motorIntensity[max_motor_index];           // What PWM value to pulse the motor with

boolean isMotorPulsing[max_motor_index];                // this tracks if the motor is currently pulsing
boolean isInRange[max_motor_index];                     // this sensor has detected we are in range
boolean isMotorScheduled[max_motor_index];              // this motor has a start time we can iterate off of since the last unscheduling

// The onboard LED (PIN 13) will alternate on and off at each sweep through the sensors to give a sense of tempo
boolean led_Value = true;  


// This method is called on time to initialize values for the motor schedules to 1 second
void initializeMotorSchedules()
{
  // Get the current time
  unsigned int currentTime = millis();

  // Initialize all motor values
  for (int x=0; x<max_motor_index; x++)
  {
    // Initialize the values for all motors
    motorStartTimeSchedule[x] = 0xFFFF;  // max value for ulong
    motorStartNextTimeSchedule[x] = 0xFFFF;  // max value for ulong
  
    motorStopTimeSchedule[x] = 0x0;  // min value for ulong
    motorStopNextTimeSchedule[x] = 0x0;  // min value for ulong
    motorStopTimeGap[x] = 0xFF;  // max value for uint
    motorIntensity[x] = 0; 
    isMotorPulsing[x] = false;   
    isInRange[x] = false;
    isMotorScheduled[x] = false;
  }
}

int pingAndAdjustSchedule(int motorToSchedule)
{ 
  // Fire the sensor
  inputSensorArray[motorToSchedule].fire();  
  int currentInchesRange = inputSensorArray[motorToSchedule].inches();
 
   // Record the current time
   unsigned long initialTime = millis();
  
  // Calculate the new gap we should have based on current range
  int new_gap = calculateRangeInFeet(currentInchesRange) * MOTOR_DISTANCE_GAP_FACTOR;
  
  // The sensors will sometime inadvertently return a range of 0.000 inches.  Disregard this reading)
  if (currentInchesRange == 0)
  {
    return 0;
  }
  
  // This event was not large enough to trigger a reschedule event
 if (abs(new_gap - motorStopTimeGap[motorToSchedule]) < RESCHEDULE_THRESHOLD_CONSTANT)
 {
    return 0;
 }


  // This event was beyond our range, let's schedule out to "end of time"
  if (new_gap >  MAX_GAP)
  {
    isInRange[motorToSchedule] = false;    
    return 0;
  }
  else // Hey, we're in range!! 
  {
    isInRange[motorToSchedule] = true;
  }  
  

  // If we are coming in range, we are scheduling our first event
  if (isMotorScheduled[motorToSchedule] == false)
  { 
    // Schedule our first motor event
    motorStartTimeSchedule[motorToSchedule] = initialTime + new_gap;
    motorStopTimeSchedule[motorToSchedule] = motorStartTimeSchedule[motorToSchedule] + MOTOR_PULSE_DURATION;

    motorStartNextTimeSchedule[motorToSchedule] = motorStopTimeSchedule[motorToSchedule] + new_gap;
    motorStopNextTimeSchedule[motorToSchedule] = motorStartNextTimeSchedule[motorToSchedule] + MOTOR_PULSE_DURATION;
    
    motorIntensity[motorToSchedule] = calculateIntensity(currentInchesRange);
    
    isMotorScheduled[motorToSchedule] = true;
    
    //DEBUG
    Serial.print("Scheduling ");
    Serial.println(motorToSchedule, DEC);
    fireAndPrintSensor(motorToSchedule, false);
    
  }
  // We are at shorter range than previously
  else if ((new_gap < motorStopTimeGap[motorToSchedule]) && (isMotorScheduled[motorToSchedule] == true))
  {
    // Adjust the new start and stop times based on the new gap value
    motorStartNextTimeSchedule[motorToSchedule] = motorStartNextTimeSchedule[motorToSchedule] - ( motorStopTimeGap[motorToSchedule] - new_gap);

    // Update our end time
    motorStopNextTimeSchedule[motorToSchedule] = motorStartNextTimeSchedule[motorToSchedule] + MOTOR_PULSE_DURATION;

    // Update the intensity   
    motorIntensity[motorToSchedule] = calculateIntensity(currentInchesRange);
   
    // Update the new value of the gap time
    motorStopTimeGap[motorToSchedule] = new_gap;
  }
  else if (new_gap > motorStopTimeGap[motorToSchedule] && (isMotorScheduled[motorToSchedule] == true))
  {
      // Adjust the new start and stop times based on the new gap value
      motorStartNextTimeSchedule[motorToSchedule] = motorStartNextTimeSchedule[motorToSchedule] + (new_gap - motorStopTimeGap[motorToSchedule]);

      // Update our end time
      motorStopNextTimeSchedule[motorToSchedule] = motorStartNextTimeSchedule[motorToSchedule] + MOTOR_PULSE_DURATION;
   
      // Update the intensity   
      motorIntensity[motorToSchedule] = calculateIntensity(currentInchesRange);
      
      // Update the new value of the gap time
      motorStopTimeGap[motorToSchedule] = new_gap;
  }
  
  return 1;
}
  
int executeMotorAndReschedule(int motorToExecute)
{
  unsigned long time = millis();
    
  // We are past the start time schedule activate the motor, our motor is pulsing, and we are scheduled
  if ((time >= motorStartTimeSchedule[motorToExecute]) && (isMotorPulsing[motorToExecute] == false) && (isMotorScheduled[motorToExecute] == true))
  {
    // Activate Motor
    // NOTE: Rather than high, this could be a PWM value to denote the intensity of the motor as well
    analogWrite(outputMotorArrayPins[motorToExecute], motorIntensity[motorToExecute]);
    //digitalWrite(outputMotorArrayPins[motorToExecute], HIGH);
    
    
    isMotorPulsing[motorToExecute] = true;    

    // Reschedule next Start Time
    motorStartTimeSchedule[motorToExecute] = motorStartNextTimeSchedule[motorToExecute];
    motorStartNextTimeSchedule[motorToExecute] = motorStartTimeSchedule[motorToExecute] + motorStopTimeGap[motorToExecute] + MOTOR_PULSE_DURATION;   

  /*
    // DEBUG:
    Serial.print("Activating Motor On Pin: ");
    Serial.println(outputMotorArrayPins[motorToExecute], DEC);
  */
    return 1;
  }  
  
  if ((time >= motorStopTimeSchedule[motorToExecute]) && (isMotorPulsing[motorToExecute] == true))
  {
    // Dectivate Motor
    analogWrite(outputMotorArrayPins[motorToExecute], 0);
    //digitalWrite(outputMotorArrayPins[motorToExecute], LOW);
    isMotorPulsing[motorToExecute] = false;        

    // We are out of range, and and we've turned off the motor.  Let's mark it as unscheduled
    if ((isInRange[motorToExecute] == false) && (isMotorScheduled[motorToExecute] == true))
    {
      isMotorScheduled[motorToExecute] = false;
      motorIntensity[motorToExecute] = 0;
        
      //DEBUG
      Serial.print("Unscheduling :");
      Serial.println(motorToExecute, DEC);
      fireAndPrintSensor(motorToExecute, false);
      
    }
    else
    {
      // Reschedule next Stop Time
      motorStopTimeSchedule[motorToExecute] = motorStopNextTimeSchedule[motorToExecute];
      motorStopNextTimeSchedule[motorToExecute] = motorStartNextTimeSchedule[motorToExecute] + MOTOR_PULSE_DURATION;   
    }

  }
  return 1;
}

void outputTest(int outputIndex)
{
  // Test all of the outputs (motors or LEDS)
    //digitalWrite(outputMotorArrayPins[outputIndex], HIGH);
    //analogWrite(outputMotorArrayPins[outputIndex], 255);
    //delay(1000);
    //digitalWrite(outputMotorArrayPins[outputIndex], LOW);
    //analogWrite(outputMotorArrayPins[outputIndex], 0);

}

int calculateIntensity(int rangeInches)
{
  // This is terrible code.  Magic numbers, and logic that could be error prone.
  if (rangeInches <=12)
  {
     return 255;
  }
  else if (rangeInches <=24)
  {
     return 128; 
     //return 50;   
  }
  else if (rangeInches <=36)
  {
     return 64;    
     //return 20;
  }
  else if (rangeInches <=48)
  {
     return 32;    
     //return 5;
  }
  else
  {
     return 0;    
  }
}

int calculateRangeInFeet(int rangeInches)
{ 
  int value = (rangeInches / 12) + 1;
  
  return value;
} 

void fireAndPrintSensor(int sensorNumber, boolean fireSensor)
{
  
  // If we have been told, then fire the sensor
  if (fireSensor == true)
  {
    // Fire the sensor
    inputSensorArray[sensorNumber].fire();  
  }  
  
  // Display the range information
  Serial.print("************ SENSOR ");
  Serial.println(sensorNumber, DEC);
  Serial.print("   Inches:");
  Serial.println(inputSensorArray[sensorNumber].inches(), DEC);
  Serial.print("   Feet:");
  Serial.println(calculateRangeInFeet(inputSensorArray[sensorNumber].inches()),DEC);
  Serial.print("   Gap in Milliseconds:");
  Serial.println(motorStopTimeGap[sensorNumber]);
  Serial.print("   Intensity:");
  Serial.println(motorIntensity[sensorNumber]);
  Serial.println();
}

void setup()
{  
  // Initialize the motors as output pins
  for (int index=left_motor;index<max_motor_index;index++)
  {
    pinMode(OUTPUT, outputMotorArrayPins[index]);
  }
                                             
  // Initialize the motors as output pins
  for (int index=left_sensor;index<max_sensor_index;index++)
  {
    pinMode(INPUT, inputSensorArrayPins[index]);
  }
  
  // Configure debug LED
  pinMode(OUTPUT_DEBUG_PIN, OUTPUT);
  
  // Initialize the serial output.  
  //****************************************************************************************
  //WARNING:  This is slow and can be hinder the performance of your system.
  //So if you gotta debug, send few characters as infrequently as possible  
  //****************************************************************************************
  Serial.begin(115200);

  // Initialize the schedules to default values;
  initializeMotorSchedules();
}

void loop()
{
 // Toggle the onboard LED to get a pace of the iterations
  digitalWrite(OUTPUT_DEBUG_PIN, led_Value);
  led_Value = !led_Value;
  
  // For each sensor, schedule it
  for (int x=0 ; x<max_sensor_index; x++)
  {  
    // DEBUG Block - Test the key functions
    //fireAndPrintSensor(x, false);
    //pingAndAdjustSchedule(2);
    //executeMotorAndReschedule(2);
    //delay(1000);
    //outputTest(x);

    // Schedule this sensor
    pingAndAdjustSchedule(x);

    // Check all motors to see if they should be run.  Note: The executeMotors function could be changed to run all motors, but 
    // this is for more control at the moment to debug individual motors
    for (int y=0; y<max_motor_index; y++)
    {
       executeMotorAndReschedule(y);
    }

  }
  // Delay for the pulse response to clear out
  delay(INTERPULSE_LATENCY_DURATION);

}
